import numpy as np
from astropy import cosmology as cosmo
from functools import wraps

from autolens import exc
from autolens.data.array import grids
from autolens.lens import plane as pl
from autolens.lens.util import lens_util
from autolens.model import cosmology_util
from autolens.model.inversion import inversions as inv
from autolens.model.inversion import pixelizations as pix
from autolens.model.galaxy import galaxy as g


def check_tracer_for_mass_profile(func):
    """If none of the tracer's galaxies have a mass profile, it surface density, potential and deflections cannot \
    be computed. This wrapper makes these properties return *None*.

    Parameters
    ----------
    func : (self) -> Object
        A property function that requires galaxies to have a mass profile.
    """

    @wraps(func)
    def wrapper(self):
        """

        Parameters
        ----------
        self

        Returns
        -------
            A value or coordinate in the same coordinate system as those passed in.
        """

        if self.has_mass_profile is True:
            return func(self)
        else:
            return None

    return wrapper


class AbstractTracerCosmology(object):

    def __init__(self, plane_redshifts, cosmology):
        """Abstract Ray tracer for lens systems with any number of planes.

        From the galaxies of the tracer's planes, their grid-stack(s) and the cosmology physically derived quantities \
        (e.g. surface density, angular diameter distances, critical surface densities) can be computed.

        Parameters
        ----------
        plane_redshifts : [pl.Plane]
            The list of the tracer's planes in ascending redshift order.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """
        self.plane_redshifts = plane_redshifts
        self.cosmology = cosmology

    @property
    def total_planes(self):
        return len(self.plane_redshifts)

    def arcsec_per_kpc_proper_of_plane(self, i):
        return cosmology_util.arcsec_per_kpc_from_redshift_and_cosmology(redshift=self.plane_redshifts[i],
                                                                         cosmology=self.cosmology)

    def kpc_per_arcsec_proper_of_plane(self, i):
        return 1.0 / self.arcsec_per_kpc_proper_of_plane(i=i)

    def angular_diameter_distance_of_plane_to_earth_in_units(self, i, unit_length='arcsec'):
        return cosmology_util.angular_diameter_distance_to_earth_from_redshift_and_cosmology(
            redshift=self.plane_redshifts[i], cosmology=self.cosmology, unit_length=unit_length)

    def angular_diameter_distance_between_planes_in_units(self, i, j, unit_length='arcsec'):
        return cosmology_util.angular_diameter_distance_between_redshifts_from_redshifts_and_cosmlology(
            redshift_0=self.plane_redshifts[i], redshift_1=self.plane_redshifts[j], cosmology=self.cosmology,
            unit_length=unit_length)

    def angular_diameter_distance_to_source_plane_in_units(self, unit_length='arcsec'):
        return cosmology_util.angular_diameter_distance_to_earth_from_redshift_and_cosmology(
            redshift=self.plane_redshifts[-1], cosmology=self.cosmology, unit_length=unit_length)

    def critical_surface_density_between_planes_in_units(self, i, j, unit_length='arcsec', unit_mass='solMass'):
        return cosmology_util.critical_surface_density_between_redshifts_from_redshifts_and_cosmology(
            redshift_0=self.plane_redshifts[i], redshift_1=self.plane_redshifts[j], cosmology=self.cosmology,
            unit_length=unit_length, unit_mass=unit_mass)

    def scaling_factor_between_planes(self, i, j):
        return cosmology_util.scaling_factor_between_redshifts_from_redshifts_and_cosmology(
            redshift_0=self.plane_redshifts[i],
            redshift_1=self.plane_redshifts[j], redshift_final=self.plane_redshifts[-1], cosmology=self.cosmology)

    def angular_diameter_distance_from_image_to_source_plane_in_units(self, unit_length='arcsec'):
        return self.angular_diameter_distance_between_planes_in_units(i=0, j=-1, unit_length=unit_length)


class AbstractTracer(AbstractTracerCosmology):

    def __init__(self, planes, cosmology):
        """Abstract Ray tracer for lens systems with any number of planes.

        From the galaxies of the tracer's planes, their grid-stack(s) and the cosmology physically derived quantities \
        (e.g. surface density, angular diameter distances, critical surface densities) can be computed.

        Parameters
        ----------
        planes : [pl.Plane]
            The list of the tracer's planes in ascending redshift order.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """

        self.planes = planes
        super(AbstractTracer, self).__init__(plane_redshifts=[plane.redshift for plane in self.planes],
                                             cosmology=cosmology)

    def plane_with_galaxy(self, galaxy):
        return [plane for plane in self.planes if galaxy in plane.galaxies][0]

    @property
    def image_plane(self):
        return self.planes[0]

    @property
    def source_plane(self):
        return self.planes[-1]

    @property
    def all_planes_have_redshifts(self):
        return None not in self.plane_redshifts

    @property
    def has_light_profile(self):
        return any(list(map(lambda plane: plane.has_light_profile, self.planes)))

    @property
    def has_mass_profile(self):
        return any(list(map(lambda plane: plane.has_mass_profile, self.planes)))

    @property
    def has_pixelization(self):
        return any(list(map(lambda plane: plane.has_pixelization, self.planes)))

    @property
    def has_regularization(self):
        return any(list(map(lambda plane: plane.has_regularization, self.planes)))

    @property
    def has_hyper_galaxy(self):
        return any(list(map(lambda plane: plane.has_hyper_galaxy, self.planes)))

    @property
    def has_padded_grids(self):
        return isinstance(self.planes[0].grids.regular, grids.PaddedRegularGrid)

    @property
    def galaxies(self):
        return list([galaxy for plane in self.planes for galaxy in plane.galaxies])

    @property
    def profile_image_plane_image_2d(self):
        return self.image_plane.grid_stack.scaled_array_2d_from_array_1d(array_1d=self.profile_image_plane_image_1d)

    @property
    def profile_image_plane_image_2d_for_simulation(self):
        return sum(self.profile_image_plane_image_2d_of_planes_for_simulation)

    @property
    def profile_image_plane_image_2d_of_planes_for_simulation(self):
        return [plane.profile_image_plane_image_2d_for_simulation for plane in self.planes]

    @property
    def profile_image_plane_image_1d(self):
        return sum(self.profile_image_plane_image_1d_of_planes)

    @property
    def profile_image_plane_image_1d_of_planes(self):
        return [plane.profile_image_plane_image_1d for plane in self.planes]

    @property
    def profile_image_plane_blurring_image_1d(self):
        return sum(self.profile_image_plane_blurring_image_1d_of_planes)

    @property
    def profile_image_plane_blurring_image_1d_of_planes(self):
        return [plane.profile_image_plane_blurring_image_1d for plane in self.planes]

    @property
    def mappers_of_planes(self):
        return list(filter(None, [plane.mapper for plane in self.planes]))

    @property
    def regularizations_of_planes(self):
        return list(filter(None, [plane.regularization for plane in self.planes]))

    @property
    @check_tracer_for_mass_profile
    def convergence(self):
        return sum([plane.convergence for plane in self.planes])

    @property
    @check_tracer_for_mass_profile
    def potential(self):
        return sum([plane.potential for plane in self.planes])

    @property
    @check_tracer_for_mass_profile
    def deflections_y(self):
        return sum([plane.deflections_y for plane in self.planes])

    @property
    @check_tracer_for_mass_profile
    def deflections_x(self):
        return sum([plane.deflections_x for plane in self.planes])

    def einstein_radius_of_plane_in_units(self, i, unit_length='arcsec'):
        return self.planes[i].einstein_radius_in_units(unit_length=unit_length)

    def einstein_mass_between_planes_in_units(self, i, j, unit_mass='solMass'):
        return self.planes[i].einstein_mass_in_units(unit_mass=unit_mass, redshift_source=self.plane_redshifts[j])

    def grid_at_redshift_from_image_plane_grid_and_redshift(self, image_plane_grid, redshift):
        """For an input grid of (y,x) arc-second image-plane coordinates, ray-trace the coordinates to any redshift in \
        the strong lens configuration.

        This is performed using multi-plane ray-tracing and the existing redshifts and planes of the tracer. However, \
        any redshift can be input even if a plane does not exist there, including redshifts before the first plane \
        of the lensing system.

        Parameters
        ----------
        image_plane_grid : ndsrray or grids.RegularGrid
            The image-plane grid which is traced to the redshift.
        redshift : float
            The redshift the image-plane grid is traced to.
        """

        # TODO : We need to come up with a better abstraction for multi-plane lensing 0_0

        image_plane_grid_stack = grids.GridStack(regular=image_plane_grid, sub=np.array([[0.0, 0.0]]),
                                                 blurring=np.array([[0.0, 0.0]]))

        tracer = TracerMultiPlanes(galaxies=self.galaxies, image_plane_grid_stack=image_plane_grid_stack,
                                   border=None, cosmology=self.cosmology)

        for plane_index in range(0, len(self.plane_redshifts)):

            new_grid_stack = image_plane_grid_stack

            if redshift <= tracer.plane_redshifts[plane_index]:

                # If redshift is between two planes, we need to map over all previous planes coordinates / deflections.

                if plane_index > 0:
                    for previous_plane_index in range(plane_index):
                        scaling_factor = cosmology_util.scaling_factor_between_redshifts_from_redshifts_and_cosmology(
                            redshift_0=tracer.plane_redshifts[previous_plane_index], redshift_1=redshift,
                            redshift_final=tracer.plane_redshifts[-1], cosmology=tracer.cosmology)

                        scaled_deflections_stack = lens_util.scaled_deflections_stack_from_plane_and_scaling_factor(
                            plane=tracer.planes[previous_plane_index], scaling_factor=scaling_factor)

                        new_grid_stack = \
                            lens_util.grid_stack_from_deflections_stack(grid_stack=new_grid_stack,
                                                                        deflections_stack=scaled_deflections_stack)

                # If redshift is before the first plane, no change to image pllane coordinates.

                elif plane_index == 0:

                    return new_grid_stack.regular

                return new_grid_stack.regular


class AbstractTracerData(AbstractTracer):

    def __init__(self, planes, cosmology):
        """Abstract Ray tracer for lens systems with any number of planes.

        From the galaxies of the tracer's planes, their grid-stack(s) and the cosmology physically derived quantities \
        (e.g. surface density, angular diameter distances, critical surface densities) can be computed.

        Parameters
        ----------
        planes : [pl.Plane]
            The list of the tracer's planes in ascending redshift order.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """
        super(AbstractTracerData, self).__init__(planes=planes, cosmology=cosmology)

    def blurred_profile_image_plane_image_2d_of_planes_from_convolver_image(self, convolver_image):
        """Extract the 1D image-plane image and 1D blurring image-plane image of every plane and blur each with the \
        PSF using a convolver (see ccd.convolution) and then map them back to the 2D array of the original mask.

        The blurred image of every plane is returned in 2D.

        Parameters
        ----------
        convolver_image : hyper.ccd.convolution.ConvolverImage
            Class which performs the PSF convolution of a masked image in 1D.
        """

        blurred_profile_image_plane_images_of_planes_1d = \
            self.blurred_profile_image_plane_image_1d_of_planes_from_convolver_image(convolver_image=convolver_image)

        return list(map(lambda blurred_profile_image_plane_image_1d:
                        self.image_plane.grid_stack.scaled_array_2d_from_array_1d(
                            array_1d=blurred_profile_image_plane_image_1d),
                        blurred_profile_image_plane_images_of_planes_1d))

    def blurred_profile_image_plane_image_1d_from_convolver_image(self, convolver_image):
        """Extract the 1D image-plane image and 1D blurring image-plane image of every plane and blur each with the \
        PSF using a convolver (see ccd.convolution).

        These are summed to give the tracer's overall blurred image-plane image in 1D.

        Parameters
        ----------
        convolver_image : hyper.ccd.convolution.ConvolverImage
            Class which performs the PSF convolution of a masked image in 1D.
        """
        return convolver_image.convolve_image(image_array=self.profile_image_plane_image_1d,
                                              blurring_array=self.profile_image_plane_blurring_image_1d)

    def blurred_profile_image_plane_image_1d_of_planes_from_convolver_image(self, convolver_image):
        """Extract the 1D image-plane image and 1D blurring image-plane image of every plane and blur each with the \
        PSF using a convolver (see ccd.convolution).

        The blurred image of every plane is returned in 1D.

        Parameters
        ----------
        convolver_image : hyper.ccd.convolution.ConvolverImage
            Class which performs the PSF convolution of a masked image in 1D.
        """
        return [plane.blurred_profile_image_plane_image_1d_from_convolver_image(convolver_image=convolver_image)
                for plane in self.planes]

    def unmasked_blurred_profile_image_plane_image_from_psf(self, psf):

        if not self.image_plane.has_padded_grid_stack:
            raise exc.RayTracingException('To retrieve an unmasked image, the grid stack of a plane tracer'
                                          'must be a padded grid stack')

        return self.image_plane.grid_stack.unmasked_blurred_image_from_psf_and_unmasked_image(
            psf=psf, unmasked_image_1d=self.profile_image_plane_image_1d)

    def unmasked_blurred_profile_image_plane_image_of_planes_from_psf(self, psf):
        return [plane.unmasked_blurred_profile_image_plane_image_from_psf(psf=psf) for plane in self.planes]

    def unmasked_blurred_profile_image_plane_image_of_plane_and_galaxies_from_psf(self, psf):
        return [plane.unmasked_blurred_profile_image_plane_images_of_galaxies_from_psf(psf=psf) for plane in
                self.planes]

    def inversion_from_image_1d_noise_map_1d_and_convolver_mapping_matrix(self, image_1d, noise_map_1d,
                                                                          convolver_mapping_matrix):

        if len(self.mappers_of_planes) > 1:
            raise exc.RayTracingException(
                'PyAutoLens does not currently support more than one mapper, reglarization and'
                'therefore inversion per tracer.')

        return inv.Inversion.from_data_1d_mapper_and_regularization(
            image_1d=image_1d, noise_map_1d=noise_map_1d, convolver=convolver_mapping_matrix,
            mapper=self.mappers_of_planes[0], regularization=self.regularizations_of_planes[0])

    def hyper_noise_map_1d_from_noise_map_1d(self, noise_map_1d):
            hyper_noise_maps_1d = self.hyper_noise_maps_1d_of_planes_from_noise_map_1d(noise_map_1d=noise_map_1d)
            hyper_noise_maps_1d = [hyper_noise_map for hyper_noise_map in hyper_noise_maps_1d if
                                   hyper_noise_map is not None]
            return sum(hyper_noise_maps_1d)

    def hyper_noise_maps_1d_of_planes_from_noise_map_1d(self, noise_map_1d):
        return [plane.hyper_noise_map_1d_from_noise_map_1d(noise_map_1d=noise_map_1d) for plane in self.planes]

    @property
    def galaxy_image_dict_blank_images(self) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """

        galaxy_image_dict = dict()

        for plane in self.planes:
            for galaxy in plane.galaxies:

                galaxy_image_dict[galaxy] = plane.profile_image_plane_image_1d_of_galaxy(galaxy=galaxy)

        return galaxy_image_dict


    def galaxy_image_dict_from_convolver_image(self, convolver_image) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """

        galaxy_image_dict = dict()

        for plane in self.planes:
            for galaxy in plane.galaxies:

                profile_image_plane_image_1d = plane.profile_image_plane_image_1d_of_galaxy(galaxy=galaxy)
                profile_image_plane_blurring_image_1d = plane.profile_image_plane_blurring_image_1d_of_galaxy(galaxy=galaxy)

                blurred_profile_image_pane_image_1d = convolver_image.convolve_image(
                    image_array=profile_image_plane_image_1d, blurring_array=profile_image_plane_blurring_image_1d)

                galaxy_image_dict[galaxy] = blurred_profile_image_pane_image_1d

        return galaxy_image_dict


class TracerImagePlane(AbstractTracerData):

    def __init__(self, lens_galaxies, image_plane_grid_stack, border=None, cosmology=cosmo.Planck15):
        """Ray tracer for a lens system with just an image-plane. 
        
        As there is only 1 plane, there are no ray-tracing calculations. This class is therefore only used for fitting \ 
        image-plane galaxies with light profiles.
        
        This tracer has only one grid-stack (see grid_stack.GridStack) which is used for ray-tracing.

        Parameters
        ----------
        lens_galaxies : [Galaxy]
            The list of lens galaxies in the image-plane.
        image_plane_grid_stack : grid_stacks.GridStack
            The image-plane grid stack which is traced. (includes the regular-grid, sub-grid, blurring-grid, etc.).
        border : masks.RegularGridBorder
            The border of the regular-grid, which is used to relocate demagnified traced pixels to the \
            source-plane borders.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """

        if not lens_galaxies:
            raise exc.RayTracingException('No lens galaxies have been input into the Tracer')

        image_plane = pl.Plane(galaxies=lens_galaxies, grid_stack=image_plane_grid_stack, compute_deflections=False,
                               border=border, cosmology=cosmology)

        super(TracerImagePlane, self).__init__(planes=[image_plane], cosmology=cosmology)

    def critical_surface_density_between_planes_in_units(self, i, j, unit_length='arcsec', unit_mass='solMass'):
        return 0.0


class TracerImageSourcePlanes(AbstractTracerData):

    def __init__(self, lens_galaxies, source_galaxies, image_plane_grid_stack, border=None, cosmology=cosmo.Planck15):
        """Ray-tracer for a lens system with two planes, an image-plane and source-plane.

        This tracer has only one grid-stack (see grid_stack.GridStack) which is used for ray-tracing.

        Parameters
        ----------
        lens_galaxies : [Galaxy]
            The list of galaxies in the image-plane.
        source_galaxies : [Galaxy]
            The list of galaxies in the source-plane.
        image_plane_grid_stack : grid_stacks.GridStack
            The image-plane grid stack which is traced. (includes the regular-grid, sub-grid, blurring-grid, etc.).
        border : masks.RegularGridBorder
            The border of the regular-grid, which is used to relocate demagnified traced pixels to the \
            source-plane borders.
        cosmology : astropy.cosmology.Planck15
            The cosmology of the ray-tracing calculation.
        """

        image_plane = pl.Plane(galaxies=lens_galaxies, grid_stack=image_plane_grid_stack, border=border,
            compute_deflections=True, cosmology=cosmology)

        source_plane_grid_stack = image_plane.trace_grid_stack_to_next_plane()

        source_plane = pl.Plane(
            galaxies=source_galaxies,
            grid_stack=source_plane_grid_stack,
            border=border,
            compute_deflections=False,
            cosmology=cosmology
        )

        super(TracerImageSourcePlanes, self).__init__(planes=[image_plane, source_plane], cosmology=cosmology)

    def critical_surface_density_between_image_and_source_plane_in_units(self, unit_length='arcsec',
                                                                         unit_mass='solMass'):
        return self.critical_surface_density_between_planes_in_units(i=0, j=1, unit_length=unit_length,
                                                                     unit_mass=unit_mass)

    def einstein_radius_of_image_plane_in_units(self, unit_length='arcsec'):
        return self.einstein_radius_of_plane_in_units(i=0, unit_length=unit_length)

    def einstein_mass_between_image_and_source_plane_in_units(self, unit_mass='solMass'):
        return self.einstein_mass_between_planes_in_units(i=0, j=1, unit_mass=unit_mass)


class TracerMultiPlanes(AbstractTracerData):

    def __init__(self, galaxies, image_plane_grid_stack, border=None, cosmology=cosmo.Planck15):
        """Ray-tracer for a lens system with any number of planes.

        The redshift of these planes are specified by the redshits of the galaxies; there is a unique plane redshift \
        for every unique galaxy redshift (galaxies with identical redshifts are put in the same plane).

        To perform multi-plane ray-tracing, a cosmology must be supplied so that deflection-angles can be rescaled \
        according to the lens-geometry of the multi-plane system. All galaxies input to the tracer must therefore \
        have redshifts.

        This tracer has only one grid-stack (see grid_stack.GridStack) which is used for ray-tracing.

        Parameters
        ----------
        galaxies : [Galaxy]
            The list of galaxies in the ray-tracing calculation.
        image_plane_grid_stack : grid_stacks.GridStack
            The image-plane grid stack which is traced. (includes the regular-grid, sub-grid, blurring-grid, etc.).
        border : masks.RegularGridBorder
            The border of the regular-grid, which is used to relocate demagnified traced pixels to the \
            source-plane borders.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """

        plane_redshifts = lens_util.ordered_plane_redshifts_from_galaxies(galaxies=galaxies)

        galaxies_in_planes = \
            lens_util.galaxies_in_redshift_ordered_planes_from_galaxies(galaxies=galaxies,
                                                                        plane_redshifts=plane_redshifts)

        planes = []

        for plane_index in range(0, len(plane_redshifts)):

            compute_deflections = lens_util.compute_deflections_at_next_plane(plane_index=plane_index,
                                                                              total_planes=len(plane_redshifts))

            new_grid_stack = image_plane_grid_stack

            if plane_index > 0:
                for previous_plane_index in range(plane_index):
                    scaling_factor = cosmology_util.scaling_factor_between_redshifts_from_redshifts_and_cosmology(
                        redshift_0=plane_redshifts[previous_plane_index], redshift_1=plane_redshifts[plane_index],
                        redshift_final=plane_redshifts[-1], cosmology=cosmology)

                    scaled_deflections_stack = lens_util.scaled_deflections_stack_from_plane_and_scaling_factor(
                        plane=planes[previous_plane_index], scaling_factor=scaling_factor)

                    new_grid_stack = \
                        lens_util.grid_stack_from_deflections_stack(grid_stack=new_grid_stack,
                                                                    deflections_stack=scaled_deflections_stack)

            planes.append(pl.Plane(galaxies=galaxies_in_planes[plane_index], grid_stack=new_grid_stack, border=border,
                                   compute_deflections=compute_deflections, cosmology=cosmology))

        super(TracerMultiPlanes, self).__init__(planes=planes, cosmology=cosmology)


class TracerMultiPlanesSliced(AbstractTracerData):

    def __init__(self, lens_galaxies, line_of_sight_galaxies, source_galaxies, planes_between_lenses,
                 image_plane_grid_stack, border=None, cosmology=cosmo.Planck15):
        """Ray-tracer for a lens system with any number of planes.

        The redshift of these planes are specified by the input parameters *lens_redshifts* and \
         *slices_between_main_planes*. Every galaxy is placed in its closest plane in redshift-space.

        To perform multi-plane ray-tracing, a cosmology must be supplied so that deflection-angles can be rescaled \
        according to the lens-geometry of the multi-plane system. All galaxies input to the tracer must therefore \
        have redshifts.

        This tracer has only one grid-stack (see grid_stack.GridStack) which is used for ray-tracing.

        Parameters
        ----------
        lens_galaxies : [Galaxy]
            The list of galaxies in the ray-tracing calculation.
        image_plane_grid_stack : grid_stacks.GridStack
            The image-plane grid stack which is traced. (includes the regular-grid, sub-grid, blurring-grid, etc.).
        planes_between_lenses : [int]
            The number of slices between each main plane. The first entry in this list determines the number of slices \
            between Earth (redshift 0.0) and main plane 0, the next between main planes 0 and 1, etc.
        border : masks.RegularGridBorder
            The border of the regular-grid, which is used to relocate demagnified traced pixels to the \
            source-plane borders.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """

        lens_redshifts = lens_util.ordered_plane_redshifts_from_galaxies(galaxies=lens_galaxies)

        plane_redshifts = lens_util.ordered_plane_redshifts_from_lens_and_source_plane_redshifts_and_slice_sizes(
            lens_redshifts=lens_redshifts, planes_between_lenses=planes_between_lenses,
            source_plane_redshift=source_galaxies[0].redshift)

        galaxies_in_planes = \
            lens_util.galaxies_in_redshift_ordered_planes_from_galaxies(galaxies=lens_galaxies + line_of_sight_galaxies,
                                                                        plane_redshifts=plane_redshifts)

        plane_redshifts.append(source_galaxies[0].redshift)
        galaxies_in_planes.append(source_galaxies)

        planes = []

        for plane_index in range(0, len(plane_redshifts)):

            compute_deflections = lens_util.compute_deflections_at_next_plane(plane_index=plane_index,
                                                                              total_planes=len(plane_redshifts))

            new_grid_stack = image_plane_grid_stack

            if plane_index > 0:
                for previous_plane_index in range(plane_index):
                    scaling_factor = cosmology_util.scaling_factor_between_redshifts_from_redshifts_and_cosmology(
                        redshift_0=plane_redshifts[previous_plane_index], redshift_1=plane_redshifts[plane_index],
                        redshift_final=plane_redshifts[-1], cosmology=cosmology)

                    scaled_deflections_stack = lens_util.scaled_deflections_stack_from_plane_and_scaling_factor(
                        plane=planes[previous_plane_index], scaling_factor=scaling_factor)

                    new_grid_stack = \
                        lens_util.grid_stack_from_deflections_stack(grid_stack=new_grid_stack,
                                                                    deflections_stack=scaled_deflections_stack)

            planes.append(pl.Plane(redshift=plane_redshifts[plane_index], galaxies=galaxies_in_planes[plane_index],
                                   grid_stack=new_grid_stack, border=border, compute_deflections=compute_deflections,
                                   cosmology=cosmology))

        super(TracerMultiPlanesSliced, self).__init__(planes=planes, cosmology=cosmology)


class TracerImageSourcePlanesPositions(AbstractTracer):

    def __init__(self, lens_galaxies, image_plane_positions, cosmology=cosmo.Planck15):
        """Positional ray-tracer for a lens system with two planes, an image-plane and source-plane (source-plane \
        galaxies are not input for the positional ray-tracer, as it is only the proximity that image_plane_positions \
        trace to within one another that needs to be computed).

        By default, this has no associated cosmology, thus all calculations are performed in arc seconds and galaxies \
        do not need input redshifts. If a cosmology is supplied, the plane's angular diameter distances, \
        conversion factors, etc. are used to provide quantities in kpc.

        Parameters
        ----------
        lens_galaxies : [Galaxy]
            The list of lens galaxies in the image-plane.
        image_plane_positions : [[[]]]
            The (y,x) arc-second coordinates of image-plane pixels which (are expected to) mappers to the same \
            location(s) in the source-plane.
        cosmology : astropy.cosmology.Planck15
            The cosmology of the ray-tracing calculation.
        """

        image_plane = pl.PlanePositions(redshift=lens_galaxies[0].redshift, galaxies=lens_galaxies,
                                        positions=image_plane_positions, compute_deflections=True, cosmology=cosmology)

        source_plane_positions = image_plane.trace_to_next_plane()

        source_plane = pl.PlanePositions(redshift=None, galaxies=None, positions=source_plane_positions,
                                         compute_deflections=False, cosmology=cosmology)

        super(TracerImageSourcePlanesPositions, self).__init__(planes=[image_plane, source_plane], cosmology=cosmology)


class TracerMultiPlanesPositions(AbstractTracer):

    def __init__(self, galaxies, image_plane_positions, cosmology=cosmo.Planck15):
        """Positional ray-tracer for a lens system with any number of planes.

        To perform multi-plane ray-tracing, a cosmology must be supplied so that deflection-angles can be rescaled \
        according to the lens-geometry of the multi-plane system.

        Parameters
        ----------
        galaxies : [Galaxy]
            The list of galaxies in the ray-tracing calculation.
        image_plane_positions : [[[]]]
            The (y,x) arc-second coordinates of image-plane pixels which (are expected to) mappers to the same \
            location(s) in the final source-plane.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """

        plane_redshifts = lens_util.ordered_plane_redshifts_from_galaxies(galaxies=galaxies)

        galaxies_in_redshift_ordered_lists = \
            lens_util.galaxies_in_redshift_ordered_planes_from_galaxies(galaxies=galaxies,
                                                                        plane_redshifts=plane_redshifts)

        if not galaxies:
            raise exc.RayTracingException('No galaxies have been input into the Tracer (TracerImageSourcePlanes)')

        planes = []

        for plane_index in range(0, len(plane_redshifts)):

            if plane_index < len(plane_redshifts) - 1:
                compute_deflections = True
            elif plane_index == len(plane_redshifts) - 1:
                compute_deflections = False
            else:
                raise exc.RayTracingException('A galaxy was not correctly allocated its previous / next redshifts')

            new_positions = image_plane_positions

            if plane_index > 0:
                for previous_plane_index in range(plane_index):
                    scaling_factor = cosmology_util.scaling_factor_between_redshifts_from_redshifts_and_cosmology(
                        redshift_0=plane_redshifts[previous_plane_index], redshift_1=plane_redshifts[plane_index],
                        redshift_final=plane_redshifts[-1], cosmology=cosmology)

                    scaled_deflections = list(map(lambda deflections:
                                                  np.multiply(scaling_factor, deflections),
                                                  planes[previous_plane_index].deflections))

                    new_positions = list(map(lambda positions, deflections:
                                             np.subtract(positions, deflections), new_positions, scaled_deflections))

            planes.append(pl.PlanePositions(redshift=plane_redshifts[plane_index],
                                            galaxies=galaxies_in_redshift_ordered_lists[plane_index],
                                            positions=new_positions, compute_deflections=compute_deflections))

        super(TracerMultiPlanesPositions, self).__init__(planes=planes, cosmology=cosmology)



import logging

from autolens import exc
from autolens.data import ccd
from autolens.data.array import grids
from autolens.data.array.scaled_array import ScaledSquarePixelArray, Array
from autolens.model.galaxy.util import galaxy_util

import numpy as np

logger = logging.getLogger(__name__)


class SimulatedCCDData(ccd.CCDData):
    
    def __init__(self, image, pixel_scale, psf, noise_map=None, background_noise_map=None, poisson_noise_map=None,
                 exposure_time_map=None, background_sky_map=None, noise_realization=None, name=None, **kwargs):
        
        super(SimulatedCCDData, self).__init__(image=image, pixel_scale=pixel_scale, psf=psf, noise_map=noise_map, 
                                               background_noise_map=background_noise_map, 
                                               poisson_noise_map=poisson_noise_map, exposure_time_map=exposure_time_map, 
                                               background_sky_map=background_sky_map, name=name, kwargs=kwargs)
        
        self.noise_realization = noise_realization

    @classmethod
    def from_deflections_source_galaxies_and_exposure_arrays(
            cls, deflections, pixel_scale, source_galaxies, exposure_time, psf=None, exposure_time_map=None,
            background_sky_level=0.0, background_sky_map=None, add_noise=True, noise_if_add_noise_false=0.1,
            noise_seed=-1, name=None):

        shape = (deflections.shape[0], deflections.shape[1])

        grid_1d = grids.RegularGrid.from_shape_and_pixel_scale(shape=shape, pixel_scale=pixel_scale)
        deflections_1d = grids.RegularGrid.from_unmasked_grid_2d(grid_2d=deflections)

        deflected_grid_1d = grid_1d - deflections_1d

        image_1d = galaxy_util.intensities_of_galaxies_from_grid(grid=deflected_grid_1d, galaxies=source_galaxies)

        image_2d = grid_1d.scaled_array_2d_from_array_1d(array_1d=image_1d)

        return cls.from_image_and_exposure_arrays(
            image=image_2d, pixel_scale=pixel_scale, exposure_time=exposure_time,
            psf=psf, exposure_time_map=exposure_time_map, background_sky_level=background_sky_level,
            background_sky_map=background_sky_map, add_noise=add_noise, noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed, name=name)

    @classmethod
    def from_tracer_and_exposure_arrays(cls, tracer, pixel_scale, exposure_time, psf=None, exposure_time_map=None,
                                       background_sky_level=0.0, background_sky_map=None, add_noise=True,
                                       noise_if_add_noise_false=0.1, noise_seed=-1, name=None):
        """
        Create a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        name
        image : ndarray
            The image before simulating (e.g. the lens and source galaxies before optics blurring and CCD read-out).
        pixel_scale: float
            The scale of each pixel in arc seconds
        exposure_time_map : ndarray
            An array representing the effective exposure time of each pixel.
        psf: PSF
            An array describing the PSF the simulated image is blurred with.
        background_sky_map : ndarray
            The value of background sky in every image pixel (electrons per second).
        add_noise: Bool
            If True poisson noise_maps is simulated and added to the image, based on the total counts in each image
            pixel
        noise_seed: int
            A seed for random noise_maps generation
        """
        return cls.from_image_and_exposure_arrays(
            image=tracer.profile_image_plane_image_2d_for_simulation, pixel_scale=pixel_scale, exposure_time=exposure_time,
            psf=psf, exposure_time_map=exposure_time_map, background_sky_level=background_sky_level,
            background_sky_map=background_sky_map, add_noise=add_noise, noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed, name=name)

    @classmethod
    def from_image_and_exposure_arrays(cls, image, pixel_scale, exposure_time, psf=None, exposure_time_map=None,
                                       background_sky_level=0.0, background_sky_map=None, add_noise=True,
                                       noise_if_add_noise_false=0.1, noise_seed=-1, name=None):
        """
        Create a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        name
        image : ndarray
            The image before simulating (e.g. the lens and source galaxies before optics blurring and CCD read-out).
        pixel_scale: float
            The scale of each pixel in arc seconds
        exposure_time_map : ndarray
            An array representing the effective exposure time of each pixel.
        psf: PSF
            An array describing the PSF the simulated image is blurred with.
        background_sky_map : ndarray
            The value of background sky in every image pixel (electrons per second).
        add_noise: Bool
            If True poisson noise_maps is simulated and added to the image, based on the total counts in each image
            pixel
        noise_seed: int
            A seed for random noise_maps generation
        """

        if psf is None:
            psf = ccd.PSF.from_no_blurring_kernel(pixel_scale=pixel_scale)
            image_needs_trimming = False
        else:
            image_needs_trimming = True

        if exposure_time_map is None:

            exposure_time_map = ScaledSquarePixelArray.single_value(value=exposure_time, shape=image.shape,
                                                                    pixel_scale=pixel_scale)

        if background_sky_map is None:

            background_sky_map = ScaledSquarePixelArray.single_value(value=background_sky_level, shape=image.shape,
                                                                     pixel_scale=pixel_scale)

        image += background_sky_map

        image = psf.convolve(image)

        if image_needs_trimming:
            image = trim_psf_edges(image, psf)
            exposure_time_map = trim_psf_edges(exposure_time_map, psf)
            background_sky_map = trim_psf_edges(background_sky_map, psf)

        if add_noise is True:
            noise_realization = generate_poisson_noise(image, exposure_time_map, noise_seed)
            image += noise_realization
            image_counts = np.multiply(image, exposure_time_map)
            noise_map = np.divide(np.sqrt(image_counts), exposure_time_map)
            noise_map = ccd.NoiseMap(array=noise_map, pixel_scale=pixel_scale)
        else:
            noise_map = ccd.NoiseMap.single_value(value=noise_if_add_noise_false, shape=image.shape,
                                                  pixel_scale=pixel_scale)

        if np.isnan(noise_map).any():
            raise exc.DataException('The noise-map has NaN values in it. This suggests your exposure time and / or'
                                       'background sky levels are too low, creating signal counts at or close to 0.0.')

        image -= background_sky_map

        # ESTIMATE THE BACKGROUND NOISE MAP FROM THE BACKGROUND SKY MAP

        background_noise_map_counts = np.sqrt(np.multiply(background_sky_map, exposure_time_map))
        background_noise_map = np.divide(background_noise_map_counts, exposure_time_map)

        # ESTIMATE THE POISSON NOISE MAP FROM THE IMAGE

        image_counts = np.multiply(image, exposure_time_map)
        poisson_noise_map = np.divide(np.sqrt(np.abs(image_counts)), exposure_time_map)

        image = ScaledSquarePixelArray(array=image, pixel_scale=pixel_scale)
        background_noise_map = ccd.NoiseMap(array=background_noise_map, pixel_scale=pixel_scale)
        poisson_noise_map = ccd.PoissonNoiseMap(array=poisson_noise_map, pixel_scale=pixel_scale)

        return SimulatedCCDData(image, pixel_scale=pixel_scale, psf=psf, noise_map=noise_map,
                                background_noise_map=background_noise_map, poisson_noise_map=poisson_noise_map,
                                exposure_time_map=exposure_time_map, background_sky_map=background_sky_map, name=name)

    def __array_finalize__(self, obj):
        if isinstance(obj, SimulatedCCDData):
            try:
                self.image = obj.image
                self.pixel_scale = obj.pixel_scale
                self.psf = obj.psf
                self.noise_map = obj.noise_map
                self.background_noise_map = obj.background_noise_map
                self.poisson_noise_map = obj.poisson_noise_map
                self.exposure_time_map = obj.exposure_time_map
                self.background_sky_map = obj.background_sky_map
                self.background_noise_realization = obj.background_noise_realization
                self.poisson_noise_realization = obj.poisson_noise_realization
                self.origin = obj.origin
            except AttributeError:
                logger.debug("Original object in CCD.__array_finalize__ missing one or more attributes")


def setup_random_seed(seed):
    """Setup the random seed. If the input seed is -1, the code will use a random seed for every run. If it is \
    positive, that seed is used for all runs, thereby giving reproducible results.

    Parameters
    ----------
    seed : int
        The seed of the random number generator.
    """
    if seed == -1:
        seed = np.random.randint(0,
                                 int(1e9))  # Use one seed, so all regions have identical column non-uniformity.
    np.random.seed(seed)


def generate_poisson_noise(image, exposure_time_map, seed=-1):
    """
    Generate a two-dimensional poisson noise_maps-mappers from an image.

    Values are computed from a Poisson distribution using the image's input values in units of counts.

    Parameters
    ----------
    image : ndarray
        The 2D image, whose values in counts are used to draw Poisson noise_maps values.
    exposure_time_map : Union(ndarray, int)
        2D array of the exposure time in each pixel used to convert to / from counts and electrons per second.
    seed : int
        The seed of the random number generator, used for the random noise_maps maps.

    Returns
    -------
    poisson_noise_map: ndarray
        An array describing simulated poisson noise_maps
    """
    setup_random_seed(seed)
    image_counts = np.multiply(image, exposure_time_map)
    return image - np.divide(np.random.poisson(image_counts, image.shape), exposure_time_map)


def trim_psf_edges(array, psf):

    if psf is not None:
        psf_cut_x = np.int(np.ceil(psf.shape[0] / 2)) - 1
        psf_cut_y = np.int(np.ceil(psf.shape[1] / 2)) - 1
        array_x = np.int(array.shape[0])
        array_y = np.int(array.shape[1])
        return array[psf_cut_x:array_x - psf_cut_x, psf_cut_y:array_y - psf_cut_y]
    else:
        return array


from autolens.lens.plotters import sensitivity_fit_plotters

from test.fixtures import *

@pytest.fixture(name='sensitivity_fit_plotter_path')
def make_sensitivity_fit_plotter_setup():
    return "{}/../../test_files/plotting/fit/".format(os.path.dirname(os.path.realpath(__file__)))


def test__fit_sub_plot__output_dependent_on_config(
        sensitivity_fit_5x5, sensitivity_fit_plotter_path, plot_patch):

    sensitivity_fit_plotters.plot_fit_subplot(
        fit=sensitivity_fit_5x5, should_plot_mask=True, extract_array_from_mask=True,
        zoom_around_mask=True, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
        output_path=sensitivity_fit_plotter_path, output_format='png')

    assert sensitivity_fit_plotter_path + 'sensitivity_fit.png' in plot_patch.paths



import numpy as np
import scipy.signal

from autolens.data import ccd
from autolens.data.array import scaled_array

class MockImage(object):

    def __new__(cls, shape, value, pixel_scale=1.0):

        return scaled_array.ScaledSquarePixelArray(array=value*np.ones(shape=shape), pixel_scale=pixel_scale)


class MockNoiseMap(object):

    def __new__(cls, shape, value, pixel_scale=1.0):

        return scaled_array.ScaledSquarePixelArray(array=value*np.ones(shape=shape), pixel_scale=pixel_scale)


class MockBackgroundNoiseMap(object):

    def __new__(cls, shape, value, pixel_scale=1.0):

        return scaled_array.ScaledSquarePixelArray(array=value*np.ones(shape=shape), pixel_scale=pixel_scale)


class MockPoissonNoiseMap(object):

    def __new__(cls, shape, value, pixel_scale=1.0):

        return scaled_array.ScaledSquarePixelArray(array=value*np.ones(shape=shape), pixel_scale=pixel_scale)

class MockExposureTimeMap(object):

    def __new__(cls, shape, value, pixel_scale=1.0):

        return scaled_array.ScaledSquarePixelArray(array=value*np.ones(shape=shape), pixel_scale=pixel_scale)


class MockBackgrondSkyMap(object):

    def __new__(cls, shape, value, pixel_scale=1.0):

        return scaled_array.ScaledSquarePixelArray(array=value*np.ones(shape=shape), pixel_scale=pixel_scale)


class MockPSF(object):

    def __new__(cls, shape, value, pixel_scale=1.0, *args, **kwargs):

        return ccd.PSF(array=value*np.ones(shape=shape), pixel_scale=pixel_scale, origin=(0.0, 0.0))


class MockImage1D(np.ndarray):

    def __new__(cls, shape, value, pixel_scale=1.0):

        array = value*np.ones(shape=shape)

        obj = np.array(array, dtype='float64').view(cls)
        obj.pixel_scale = pixel_scale
        obj.origin = (0.0, 0.0)

        return obj


class MockNoiseMap1D(np.ndarray):

    def __new__(cls, shape, value, pixel_scale=1.0):

        array = value*np.ones(shape=shape)

        obj = np.array(array, dtype='float64').view(cls)
        obj.pixel_scale = pixel_scale
        obj.origin = (0.0, 0.0)

        return obj


class MockCCDData(ccd.CCDData):

    def __init__(self, image, pixel_scale, psf, noise_map, background_noise_map, poisson_noise_map,
                 exposure_time_map, background_sky_map, name):

        super(MockCCDData, self).__init__(
            image=image, pixel_scale=pixel_scale, psf=psf, noise_map=noise_map,
            background_noise_map=background_noise_map, poisson_noise_map=poisson_noise_map,
            exposure_time_map=exposure_time_map, background_sky_map=background_sky_map, name=name)


import numpy as np
import pytest

from autolens.data import convolution
from autolens.data.array import mask as msk
from autolens.lens.util import lens_fit_util as util
from autolens.model.galaxy import galaxy as g
from autolens.model.profiles import light_profiles as lp


@pytest.fixture(name='mask')
def make_mask():
    return msk.Mask(array=np.array([[True, True, True, True],
                                    [True, False, False, True],
                                    [True, False, False, True],
                                    [True, True, True, True]]), pixel_scale=1.0)


@pytest.fixture(name='blurring_mask')
def make_blurring_mask():
    return msk.Mask(array=np.array([[False, False, False, False],
                                    [False, True, True, False],
                                    [False, True, True, False],
                                    [False, False, False, False]]), pixel_scale=1.0)


@pytest.fixture(name='convolver_no_blur')
def make_convolver_no_blur(mask, blurring_mask):
    psf = np.array([[0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0]])

    return convolution.ConvolverImage(mask=mask, blurring_mask=blurring_mask, psf=psf)


@pytest.fixture(name='convolver_blur')
def make_convolver_blur(mask, blurring_mask):
    psf = np.array([[1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0]])

    return convolution.ConvolverImage(mask=mask, blurring_mask=blurring_mask, psf=psf)


@pytest.fixture(name="galaxy_light")
def make_galaxy_light():
    return g.Galaxy(redshift=0.5, light_profile=lp.EllipticalSersic(centre=(0.1, 0.1), axis_ratio=1.0, phi=0.0, intensity=1.0,
                                                      effective_radius=0.6, sersic_index=4.0))


class TestInversionEvidence:

    def test__simple_values(self):
        likelihood_with_regularization_terms = \
            util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                chi_squared=3.0, regularization_term=6.0, noise_normalization=2.0)

        assert likelihood_with_regularization_terms == -0.5 * (3.0 + 6.0 + 2.0)

        evidences = util.evidence_from_inversion_terms(chi_squared=3.0, regularization_term=6.0,
                                                       log_curvature_regularization_term=9.0,
                                                       log_regularization_term=10.0, noise_normalization=30.0)

        assert evidences == -0.5 * (3.0 + 6.0 + 9.0 - 10.0 + 30.0)


# class TestUnmaskedModelImage:
#
#     def test___3x3_padded_image__asymmetric_psf_blurring__produces_padded_image(self):
#         mask = msk.Mask(array=np.array([[True, True, True],
#                                         [True, False, True],
#                                         [True, True, True]]), pixel_scale=1.0)
#
#         psf = im.PSF(array=(np.array([[0.0, 3.0, 0.0],
#                                       [0.0, 1.0, 2.0],
#                                       [0.0, 0.0, 0.0]])), pixel_scale=1.0)
#
#         padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
#                                                                                                     sub_grid_size=1,
#                                                                                                     psf_shape=(3, 3))
#
#         unmasked_image_1d = np.zeros(25)
#         unmasked_image_1d[12] = 1.0
#
#         unmasked_blurred_image = padded_grid_stack.unmasked_blurred_image_from_psf_and_unmasked_image(
#             psf=psf, unmasked_image_1d=unmasked_image_1d)
#
#         assert (unmasked_blurred_image == np.array([[0.0, 3.0, 0.0],
#                                                     [0.0, 1.0, 2.0],
#                                                     [0.0, 0.0, 0.0]])).all()


import numpy as np

from autolens.model.galaxy.plotters import galaxy_fit_plotters
from test.fixtures import *

import numpy as np

@pytest.fixture(name='galaxy_fitting_plotter_path')
def make_galaxy_fitting_plotter_setup():
    return "{}/../../../test_files/plotting/galaxy_fitting/".format(os.path.dirname(os.path.realpath(__file__)))


def test__fit_sub_plot__galaxy_intensities(gal_fit_5x5_intensities, positions_5x5, plot_patch,
                                           galaxy_fitting_plotter_path):

    galaxy_fit_plotters.plot_fit_subplot(fit=gal_fit_5x5_intensities,
                                         should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
                                         positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                         output_path=galaxy_fitting_plotter_path, output_format='png')

    assert galaxy_fitting_plotter_path + 'galaxy_fit.png' in plot_patch.paths


def test__fit_sub_plot__galaxy_convergence(gal_fit_5x5_convergence, mask_5x5, positions_5x5,
                                           plot_patch,
                                           galaxy_fitting_plotter_path):

    galaxy_fit_plotters.plot_fit_subplot(fit=gal_fit_5x5_convergence,
                                         should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
                                         positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                         output_path=galaxy_fitting_plotter_path, output_format='png')

    assert galaxy_fitting_plotter_path + 'galaxy_fit.png' in plot_patch.paths


def test__fit_sub_plot__galaxy_potential(gal_fit_5x5_potential, mask_5x5, positions_5x5, plot_patch,
                                         galaxy_fitting_plotter_path):

    galaxy_fit_plotters.plot_fit_subplot(fit=gal_fit_5x5_potential,
                                         should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
                                         positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                         output_path=galaxy_fitting_plotter_path, output_format='png')

    assert galaxy_fitting_plotter_path + 'galaxy_fit.png' in plot_patch.paths


def test__fit_sub_plot__galaxy_deflections_y_(gal_fit_5x5_deflections_y, mask_5x5, positions_5x5, plot_patch,
                                              galaxy_fitting_plotter_path):

    galaxy_fit_plotters.plot_fit_subplot(fit=gal_fit_5x5_deflections_y,
                                         should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
                                         positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                         output_path=galaxy_fitting_plotter_path, output_format='png')

    assert galaxy_fitting_plotter_path + 'galaxy_fit.png' in plot_patch.paths


def test__fit_sub_plot__galaxy_deflections_x(gal_fit_5x5_deflections_x, mask_5x5, positions_5x5, plot_patch,
                                             galaxy_fitting_plotter_path):

    galaxy_fit_plotters.plot_fit_subplot(fit=gal_fit_5x5_deflections_x,
                                         should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
                                         positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                         output_path=galaxy_fitting_plotter_path, output_format='png')

    assert galaxy_fitting_plotter_path + 'galaxy_fit.png' in plot_patch.paths



import numpy as np
import pytest

from autolens.data import convolution
from autolens.data.array.util import grid_util
from autolens.data import ccd
from autolens.data.array import grids
from autolens.data.array import mask as msk
from autolens.lens import lens_data as ld
from autolens.model.inversion import convolution as inversion_convolution

@pytest.fixture(name="lens_data_5x5")
def make_lens_data_5x5(ccd_data_5x5, mask_5x5):
    return ld.LensData(ccd_data=ccd_data_5x5, mask=mask_5x5)

@pytest.fixture(name="lens_data_6x6")
def make_lens_data_6x6(ccd_data_6x6, mask_6x6):
    return ld.LensData(ccd_data=ccd_data_6x6, mask=mask_6x6)

class TestLensData(object):

    def test__attributes(
            self, ccd_data_5x5, lens_data_5x5):

        assert lens_data_5x5.pixel_scale == ccd_data_5x5.pixel_scale
        assert lens_data_5x5.pixel_scale == 1.0

        assert (lens_data_5x5.unmasked_image == ccd_data_5x5.image).all()
        assert (lens_data_5x5.unmasked_image == np.ones((5,5))).all()

        assert (lens_data_5x5.unmasked_noise_map == ccd_data_5x5.noise_map).all()
        assert (lens_data_5x5.unmasked_noise_map == 2.0*np.ones((5,5))).all()

        assert (lens_data_5x5.psf == ccd_data_5x5.psf).all()
        assert (lens_data_5x5.psf == np.ones((3,3))).all()

        assert lens_data_5x5.image_psf_shape == (3,3)
        assert lens_data_5x5.inversion_psf_shape == (3,3)

    def test__masking(
            self, lens_data_5x5):

        assert (lens_data_5x5.mask_1d == np.full(fill_value=False, shape=(9))).all()
        assert (lens_data_5x5.image_1d == np.ones(9)).all()
        assert (lens_data_5x5.noise_map_1d == 2.0*np.ones(9)).all()

        assert (lens_data_5x5.mask_2d == np.array([[True, True,  True,  True, True],
                                              [True, False, False, False, True],
                                              [True, False, False, False, True],
                                              [True, False, False, False, True],
                                              [True,  True,  True,  True, True]])).all()

        assert (lens_data_5x5.image_2d == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                [0.0, 1.0, 1.0, 1.0, 0.0],
                                                [0.0, 1.0, 1.0, 1.0, 0.0],
                                                [0.0, 1.0, 1.0, 1.0, 0.0],
                                                [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

        assert (lens_data_5x5.noise_map_2d == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

    def test__grid_stack(
            self, lens_data_5x5, regular_grid_5x5, sub_grid_5x5, blurring_grid_5x5):

        assert (lens_data_5x5.grid_stack.regular == regular_grid_5x5).all()
        assert (lens_data_5x5.grid_stack.sub == sub_grid_5x5).all()
        assert (lens_data_5x5.grid_stack.blurring == blurring_grid_5x5).all()

    def test__padded_grid_stack(
            self, lens_data_5x5):

        padded_image_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(
            mask=np.full((7, 7), False), pixel_scales=lens_data_5x5.unmasked_image.pixel_scales)

        assert (lens_data_5x5.padded_grid_stack.regular == padded_image_util).all()
        assert lens_data_5x5.padded_grid_stack.regular.image_shape == (5, 5)
        assert lens_data_5x5.padded_grid_stack.regular.padded_shape == (7, 7)

        padded_sub_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=np.full((7, 7), False), pixel_scales=lens_data_5x5.unmasked_image.pixel_scales,
            sub_grid_size=lens_data_5x5.grid_stack.sub.sub_grid_size)

        assert lens_data_5x5.padded_grid_stack.sub == pytest.approx(padded_sub_util, 1e-4)
        assert lens_data_5x5.padded_grid_stack.sub.image_shape == (5, 5)
        assert lens_data_5x5.padded_grid_stack.sub.padded_shape == (7, 7)

        assert (lens_data_5x5.padded_grid_stack.blurring == np.array([[0.0, 0.0]])).all()

    def test__interp_pixel_scale_input__grid_stack_and_padded_grid_stack_include_interpolators(
            self, ccd_data_5x5, mask_5x5):

        lens_data_5x5 = ld.LensData(ccd_data=ccd_data_5x5, mask=mask_5x5, interp_pixel_scale=1.0)

        grid_stack = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(
                        mask=mask_5x5, sub_grid_size=2, psf_shape=(3, 3))
        new_grid_stack = grid_stack.new_grid_stack_with_interpolator_added_to_each_grid(interp_pixel_scale=1.0)

        assert (lens_data_5x5.grid_stack.regular == new_grid_stack.regular).all()
        assert (lens_data_5x5.grid_stack.regular.interpolator.vtx == new_grid_stack.regular.interpolator.vtx).all()
        assert (lens_data_5x5.grid_stack.regular.interpolator.wts == new_grid_stack.regular.interpolator.wts).all()

        assert (lens_data_5x5.grid_stack.sub == new_grid_stack.sub).all()
        assert (lens_data_5x5.grid_stack.sub.interpolator.vtx == new_grid_stack.sub.interpolator.vtx).all()
        assert (lens_data_5x5.grid_stack.sub.interpolator.wts == new_grid_stack.sub.interpolator.wts).all()
        
        assert (lens_data_5x5.grid_stack.blurring == new_grid_stack.blurring).all()
        assert (lens_data_5x5.grid_stack.blurring.interpolator.vtx == new_grid_stack.blurring.interpolator.vtx).all()
        assert (lens_data_5x5.grid_stack.blurring.interpolator.wts == new_grid_stack.blurring.interpolator.wts).all()

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(
                        mask=mask_5x5, sub_grid_size=2, psf_shape=(3, 3))
        new_padded_grid_stack = \
            padded_grid_stack.new_grid_stack_with_interpolator_added_to_each_grid(interp_pixel_scale=1.0)

        assert (lens_data_5x5.padded_grid_stack.regular == new_padded_grid_stack.regular).all()
        assert (lens_data_5x5.padded_grid_stack.regular.interpolator.vtx == new_padded_grid_stack.regular.interpolator.vtx).all()
        assert (lens_data_5x5.padded_grid_stack.regular.interpolator.wts == new_padded_grid_stack.regular.interpolator.wts).all()

        assert (lens_data_5x5.padded_grid_stack.sub == new_padded_grid_stack.sub).all()
        assert (lens_data_5x5.padded_grid_stack.sub.interpolator.vtx == new_padded_grid_stack.sub.interpolator.vtx).all()
        assert (lens_data_5x5.padded_grid_stack.sub.interpolator.wts == new_padded_grid_stack.sub.interpolator.wts).all()

        assert (lens_data_5x5.padded_grid_stack.blurring == np.array([[0.0, 0.0]])).all()

    def test__cluster_pixel_scale_is_input__correct_cluster_bin_up_calculated__inversion_max_pixels_changes_bin_up(
            self, ccd_data_5x5, mask_5x5, regular_grid_5x5):

        ccd_data_5x5.pixel_scale = 1.0
        lens_data_5x5 = ld.LensData(
            ccd_data=ccd_data_5x5, mask=mask_5x5, cluster_pixel_scale=1.0)

        assert lens_data_5x5.cluster.bin_up_factor == 1
        assert (lens_data_5x5.mask_2d == lens_data_5x5.cluster.mask).all()
        assert (lens_data_5x5.cluster == regular_grid_5x5).all()
        assert (lens_data_5x5.cluster.cluster_to_regular_all == np.array(
            [[0], [1], [2], [3], [4], [5], [6], [7], [8]])).all()

        ccd_data_5x5.pixel_scale = 1.0
        lens_data_5x5 = ld.LensData(
            ccd_data=ccd_data_5x5, mask=mask_5x5, cluster_pixel_scale=1.9)

        assert lens_data_5x5.cluster.bin_up_factor == 1
        assert (lens_data_5x5.mask_2d == lens_data_5x5.cluster.mask).all()
        assert (lens_data_5x5.cluster.cluster_to_regular_all == np.array(
            [[0], [1], [2], [3], [4], [5], [6], [7], [8]])).all()

        ccd_data_5x5.pixel_scale = 1.0
        lens_data_5x5 = ld.LensData(
            ccd_data=ccd_data_5x5, mask=mask_5x5, cluster_pixel_scale=2.0)
        assert lens_data_5x5.cluster.bin_up_factor == 2
        assert (lens_data_5x5.cluster.mask == np.array([[True, True, True],
                                                           [True, False, False],
                                                           [True, False, False]])).all()
        assert (lens_data_5x5.cluster == np.array(
            [[0.0, 0.0], [0.0, 2.0], [-2.0, 0.0], [-2.0, 2.0]])).all()
        assert (lens_data_5x5.cluster.cluster_to_regular_all == np.array(
            [[0, 1, 3, 4],
            [2, 5, -1, -1],
             [6, 7, -1, -1],
             [8, -1, -1, -1]])).all()

        ccd_data_5x5.pixel_scale = 2.0
        lens_data_5x5 = ld.LensData(
            ccd_data=ccd_data_5x5, mask=mask_5x5, cluster_pixel_scale=1.0)

        assert lens_data_5x5.cluster.bin_up_factor == 1

        ccd_data_5x5.pixel_scale = 1.0
        lens_data_5x5 = ld.LensData(
            ccd_data=ccd_data_5x5, mask=mask_5x5, cluster_pixel_scale=None)

        assert lens_data_5x5.cluster.bin_up_factor == 1
        assert (lens_data_5x5.mask_2d == lens_data_5x5.cluster.mask).all()
        assert (lens_data_5x5.cluster == regular_grid_5x5).all()
        assert (lens_data_5x5.cluster.cluster_to_regular_all == np.array(
            [[0], [1], [2], [3], [4], [5], [6], [7], [8]])).all()

        ccd_data_5x5.pixel_scale = 1.0
        lens_data_5x5 = ld.LensData(
            ccd_data=ccd_data_5x5, mask=mask_5x5, cluster_pixel_scale=2.0, cluster_pixel_limit=5)

        assert lens_data_5x5.cluster.bin_up_factor == 1
        assert (lens_data_5x5.mask_2d == lens_data_5x5.cluster.mask).all()
        assert (lens_data_5x5.cluster.cluster_to_regular_all == np.array(
            [[0], [1], [2], [3], [4], [5], [6], [7], [8]])).all()


    def test__border(
            self, lens_data_5x5):

        assert (lens_data_5x5.border == np.array([0, 1, 2, 3, 5, 6, 7, 8])).all()

    def test__convolvers(
            self, lens_data_5x5):
        assert type(lens_data_5x5.convolver_image) == convolution.ConvolverImage
        assert type(lens_data_5x5.convolver_mapping_matrix) == inversion_convolution.ConvolverMappingMatrix

    def test__inversion_max_pixels(
            self, ccd_data_5x5, mask_5x5):

        lens_data_5x5 = ld.LensData(
            ccd_data=ccd_data_5x5, mask=mask_5x5, cluster_pixel_limit=2)

        assert lens_data_5x5.cluster_pixel_limit == 2

        lens_data_5x5 = ld.LensData(
            ccd_data=ccd_data_5x5, mask=mask_5x5, cluster_pixel_limit=5)

        assert lens_data_5x5.cluster_pixel_limit == 5

    def test__uses_inversion__does_not_create_mapping_matrix_conovolver_if_false(
            self, ccd_data_5x5, mask_5x5):

        lens_data_5x5 = ld.LensData(ccd_data=ccd_data_5x5, mask=mask_5x5, uses_inversion=False)

        assert lens_data_5x5.convolver_mapping_matrix == None

    def test__uses_cluster_inversion__does_not_create_cluster_grid_if_false(
            self, ccd_data_5x5, mask_5x5):

        lens_data_5x5 = ld.LensData(
            ccd_data=ccd_data_5x5, mask=mask_5x5,
            cluster_pixel_scale=1.0, cluster_pixel_limit=1, uses_cluster_inversion=False)

        assert lens_data_5x5.cluster == None
        assert lens_data_5x5.cluster_pixel_scale == None
        assert lens_data_5x5.cluster_pixel_limit == None

    def test__different_ccd_data_without_mock_objects__customize_constructor_inputs(self):

        psf = ccd.PSF(np.ones((7, 7)), 1)
        ccd_data = ccd.CCDData(np.ones((19, 19)), pixel_scale=3., psf=psf, noise_map=2.0*np.ones((19, 19)))
        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(19, 19), pixel_scale=1.0, invert=True)
        mask[9, 9] = False

        lens_data_5x5 = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=8, image_psf_shape=(5, 5),
                                inversion_psf_shape=(3, 3), positions=[np.array([[1.0, 1.0]])])

        assert (lens_data_5x5.unmasked_image == np.ones((19, 19))).all()
        assert (lens_data_5x5.unmasked_noise_map == 2.0*np.ones((19, 19))).all()
        assert (lens_data_5x5.psf == np.ones((7,7))).all()

        assert lens_data_5x5.sub_grid_size == 8
        assert lens_data_5x5.convolver_image.psf_shape == (5, 5)
        assert lens_data_5x5.convolver_mapping_matrix.psf_shape == (3, 3)
        assert (lens_data_5x5.positions[0] == np.array([[1.0, 1.0]])).all()

        assert lens_data_5x5.image_psf_shape == (5,5)
        assert lens_data_5x5.inversion_psf_shape == (3, 3)

    def test__lens_data_5x5_with_modified_image(
            self, lens_data_5x5):

        lens_data_5x5 = lens_data_5x5.new_lens_data_with_modified_image(modified_image=8.0 * np.ones((5, 5)))

        assert (lens_data_5x5.unmasked_image == 8.0*np.ones((5,5))).all()

        assert (lens_data_5x5.image_1d == 8.0*np.ones(9)).all()

        assert (lens_data_5x5.image_2d == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                [0.0, 8.0, 8.0, 8.0, 0.0],
                                                [0.0, 8.0, 8.0, 8.0, 0.0],
                                                [0.0, 8.0, 8.0, 8.0, 0.0],
                                                [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

    def test__lens_data_6x6_with_binned_up_ccd_data(
            self, lens_data_6x6):

        binned_up_psf = lens_data_6x6.ccd_data.psf.new_psf_with_rescaled_odd_dimensioned_array(rescale_factor=0.5)

        lens_data_6x6 = lens_data_6x6.new_lens_data_with_binned_up_ccd_data_and_mask(bin_up_factor=2)

        assert (lens_data_6x6.unmasked_image == np.ones((3,3))).all()
        assert (lens_data_6x6.psf == binned_up_psf).all()
        assert (lens_data_6x6.unmasked_noise_map == np.ones((3,3))).all()
        assert (lens_data_6x6.ccd_data.background_noise_map == 1.5*np.ones((3,3))).all()
        assert (lens_data_6x6.ccd_data.poisson_noise_map == 2.0*np.ones((3,3))).all()
        assert (lens_data_6x6.ccd_data.exposure_time_map == 20.0*np.ones((3,3))).all()
        assert (lens_data_6x6.ccd_data.background_sky_map == 6.0*np.ones((3,3))).all()

        assert (lens_data_6x6.mask_2d == np.array([[True, True, True],
                                               [True, False, True],
                                               [True, True, True]])).all()

        assert (lens_data_6x6.image_1d == np.ones((1))).all()
        assert (lens_data_6x6.noise_map_1d == np.ones((1))).all()


import numpy as np
import pytest

import autofit as af
from autolens.data import ccd
from autolens.data.array import mask as mask
from autolens.lens import lens_data as ld
from autolens.lens import ray_tracing
from autolens.lens import sensitivity_fit
from autolens.model.galaxy import galaxy as g
from autolens.model.inversion import inversions as inv
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp


@pytest.fixture(name="sersic")
def make_sersic():
    return lp.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0,
                               effective_radius=0.6, sersic_index=4.0)


@pytest.fixture(name="galaxy_light", scope='function')
def make_galaxy_light(sersic):
    return g.Galaxy(redshift=0.5, light_profile=sersic)


@pytest.fixture(name='ld_blur')
def make_ld_blur():
    psf = ccd.PSF(array=(np.array([[1.0, 1.0, 1.0],
                                   [1.0, 1.0, 1.0],
                                   [1.0, 1.0, 1.0]])), pixel_scale=1.0,
                  renormalize=False)
    im = ccd.CCDData(image=5.0 * np.ones((6, 6)), pixel_scale=1.0, psf=psf,
                     noise_map=np.ones((6, 6)),
                     exposure_time_map=3.0 * np.ones((6, 6)),
                     background_sky_map=4.0 * np.ones((6, 6)))

    ma = np.array([[True, True, True, True, True, True],
                   [True, False, False, False, False, True],
                   [True, False, False, False, False, True],
                   [True, False, False, False, False, True],
                   [True, False, False, False, False, True],
                   [True, True, True, True, True, True]])
    ma = mask.Mask(array=ma, pixel_scale=1.0)

    return ld.LensData(ccd_data=im, mask=ma, sub_grid_size=2)


class TestSensitivityProfileFit:

    def test__tracer_and_tracer_sensitive_are_identical__added__likelihood_is_noise_term(
            self, ld_blur):
        g0 = g.Galaxy(redshift=0.5,
                      mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))
        g1 = g.Galaxy(redshift=1.0, light_profile=lp.EllipticalSersic(intensity=2.0))

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0],
                                                     source_galaxies=[g1],
                                                     image_plane_grid_stack=ld_blur.grid_stack)

        fit = sensitivity_fit.SensitivityProfileFit(lens_data=ld_blur,
                                                    tracer_normal=tracer,
                                                    tracer_sensitive=tracer)

        assert (fit.fit_normal.image_2d == ld_blur.image_2d).all()
        assert (fit.fit_normal.noise_map_2d == ld_blur.noise_map_2d).all()

        model_image_1d = tracer.blurred_profile_image_plane_image_1d_from_convolver_image(
            convolver_image=ld_blur.convolver_image)

        model_image_2d = ld_blur.map_to_scaled_array(array_1d=model_image_1d)

        assert (fit.fit_normal.model_image_2d == model_image_2d).all()

        residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=ld_blur.image_1d, mask=ld_blur.mask_1d, model_data=model_image_1d)

        residual_map_2d = ld_blur.map_to_scaled_array(array_1d=residual_map_1d)
        assert (fit.fit_normal.residual_map_2d == residual_map_2d).all()

        chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_2d,
            mask=ld_blur.mask_2d, noise_map=ld_blur.noise_map_2d)

        assert (fit.fit_normal.chi_squared_map_2d == chi_squared_map_2d).all()

        assert (fit.fit_sensitive.image_2d == ld_blur.image_2d).all()
        assert (fit.fit_sensitive.noise_map_2d == ld_blur.noise_map_2d).all()
        assert (fit.fit_sensitive.model_image_2d == model_image_2d).all()
        assert (fit.fit_sensitive.residual_map_2d == residual_map_2d).all()
        assert (fit.fit_sensitive.chi_squared_map_2d == chi_squared_map_2d).all()

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=chi_squared_map_2d, mask=ld_blur.mask_2d)
        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            mask=ld_blur.mask_2d, noise_map=ld_blur.noise_map_2d)

        assert fit.fit_normal.likelihood == -0.5 * (chi_squared + noise_normalization)
        assert fit.fit_sensitive.likelihood == -0.5 * (
                chi_squared + noise_normalization)

        assert fit.figure_of_merit == 0.0

    def test__tracers_are_different__likelihood_is_non_zero(self, ld_blur):
        g0 = g.Galaxy(redshift=0.5,
                      mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))
        g0_subhalo = g.Galaxy(redshift=0.5,
                              subhalo=mp.SphericalIsothermal(einstein_radius=0.1))
        g1 = g.Galaxy(redshift=1.0, light_profile=lp.EllipticalSersic(intensity=2.0))

        tracer_normal = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0],
                                                            source_galaxies=[g1],
                                                            image_plane_grid_stack=ld_blur.grid_stack)

        tracer_sensitive = ray_tracing.TracerImageSourcePlanes(
            lens_galaxies=[g0, g0_subhalo], source_galaxies=[g1],
            image_plane_grid_stack=ld_blur.grid_stack)

        fit = sensitivity_fit.SensitivityProfileFit(lens_data=ld_blur,
                                                    tracer_normal=tracer_normal,
                                                    tracer_sensitive=tracer_sensitive)

        assert (fit.fit_normal.image_2d == ld_blur.image_2d).all()
        assert (fit.fit_normal.noise_map_2d == ld_blur.noise_map_2d).all()

        model_image_1d = tracer_normal.blurred_profile_image_plane_image_1d_from_convolver_image(
            convolver_image=ld_blur.convolver_image)

        model_image_2d = ld_blur.map_to_scaled_array(array_1d=model_image_1d)

        assert (fit.fit_normal.model_image_2d == model_image_2d).all()

        residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=ld_blur.image_2d, mask=ld_blur.mask_2d, model_data=model_image_2d)

        assert (fit.fit_normal.residual_map_2d == residual_map_2d).all()

        chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_2d, mask=ld_blur.mask_2d,
            noise_map=ld_blur.noise_map_2d)

        assert (fit.fit_normal.chi_squared_map_2d == chi_squared_map_2d).all()

        assert (fit.fit_sensitive.image_2d == ld_blur.image_2d).all()
        assert (fit.fit_sensitive.noise_map_2d == ld_blur.noise_map_2d).all()

        model_image_1d = tracer_sensitive.blurred_profile_image_plane_image_1d_from_convolver_image(
            convolver_image=ld_blur.convolver_image)

        model_image_2d = ld_blur.map_to_scaled_array(array_1d=model_image_1d)

        assert (fit.fit_sensitive.model_image_2d == model_image_2d).all()

        residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=ld_blur.image_2d, mask=ld_blur.mask_2d, model_data=model_image_2d)

        assert (fit.fit_sensitive.residual_map_2d == residual_map_2d).all()

        chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_2d, mask=ld_blur.mask_2d,
            noise_map=ld_blur.noise_map_2d)

        assert (fit.fit_sensitive.chi_squared_map_2d == chi_squared_map_2d).all()

        chi_squared_normal = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=fit.fit_normal.chi_squared_map_2d, mask=ld_blur.mask_2d)
        chi_squared_sensitive = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=fit.fit_sensitive.chi_squared_map_2d, mask=ld_blur.mask_2d)
        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            mask=ld_blur.mask_2d, noise_map=ld_blur.noise_map_2d)

        assert fit.fit_normal.likelihood == -0.5 * (
                chi_squared_normal + noise_normalization)
        assert fit.fit_sensitive.likelihood == -0.5 * (
                chi_squared_sensitive + noise_normalization)

        assert fit.figure_of_merit == fit.fit_sensitive.likelihood - fit.fit_normal.likelihood

        fit_from_factory = sensitivity_fit.fit_lens_data_with_sensitivity_tracers(
            lens_data=ld_blur,
            tracer_normal=tracer_normal,
            tracer_sensitive=tracer_sensitive)

        assert fit.figure_of_merit == fit_from_factory.figure_of_merit


class TestSensitivityInversionFit:

    def test__tracer_and_tracer_sensitive_are_identical__added__likelihood_is_noise_term(
            self, ld_blur):
        pixelization = pix.Rectangular(shape=(3, 3))
        regularization = reg.Constant(coefficients=(1.0,))

        g0 = g.Galaxy(redshift=0.5,
                      mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))
        g1 = g.Galaxy(redshift=1.0, pixelization=pixelization,
                      regularization=regularization)

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0],
                                                     source_galaxies=[g1],
                                                     image_plane_grid_stack=ld_blur.grid_stack)

        fit = sensitivity_fit.SensitivityInversionFit(lens_data=ld_blur,
                                                      tracer_normal=tracer,
                                                      tracer_sensitive=tracer)

        assert (fit.fit_normal.image_2d == ld_blur.image_2d).all()
        assert (fit.fit_normal.noise_map_2d == ld_blur.noise_map_2d).all()

        mapper = pixelization.mapper_from_grid_stack_and_border(
            grid_stack=tracer.source_plane.grid_stack, border=None)
        inversion = inv.Inversion.from_data_1d_mapper_and_regularization(
            mapper=mapper, regularization=regularization, image_1d=ld_blur.image_1d,
            noise_map_1d=ld_blur.noise_map_1d,
            convolver=ld_blur.convolver_mapping_matrix)

        assert fit.fit_normal.model_image_2d == pytest.approx(
            inversion.reconstructed_data_2d, 1.0e-4)

        residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=ld_blur.image_2d, mask=ld_blur.mask_2d,
            model_data=inversion.reconstructed_data_2d)

        assert fit.fit_normal.residual_map_2d == pytest.approx(residual_map_2d, 1.0e-4)

        chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_2d, mask=ld_blur.mask_2d,
            noise_map=ld_blur.noise_map_2d)

        assert fit.fit_normal.chi_squared_map_2d == pytest.approx(chi_squared_map_2d,
                                                                  1.0e-4)

        assert fit.fit_sensitive.image_2d == pytest.approx(ld_blur.image_2d, 1.0e-4)
        assert fit.fit_sensitive.noise_map_2d == pytest.approx(ld_blur.noise_map_2d,
                                                               1.0e-4)
        assert fit.fit_sensitive.model_image_2d == pytest.approx(
            inversion.reconstructed_data_2d, 1.0e-4)
        assert fit.fit_sensitive.residual_map_2d == pytest.approx(residual_map_2d,
                                                                  1.0e-4)
        assert fit.fit_sensitive.chi_squared_map_2d == pytest.approx(chi_squared_map_2d,
                                                                     1.0e-4)

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=chi_squared_map_2d,
            mask=ld_blur.mask_2d)
        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            mask=ld_blur.mask_2d,
            noise_map=ld_blur.noise_map_2d)

        assert fit.fit_normal.likelihood == -0.5 * (chi_squared + noise_normalization)
        assert fit.fit_sensitive.likelihood == -0.5 * (
                chi_squared + noise_normalization)

        assert fit.figure_of_merit == 0.0

    def test__tracers_are_different__likelihood_is_non_zero(self, ld_blur):
        pixelization = pix.Rectangular(shape=(3, 3))
        regularization = reg.Constant(coefficients=(1.0,))

        g0 = g.Galaxy(redshift=0.5,
                      mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))
        g0_subhalo = g.Galaxy(redshift=0.5,
                              subhalo=mp.SphericalIsothermal(einstein_radius=0.1))
        g1 = g.Galaxy(redshift=1.0, pixelization=pixelization,
                      regularization=regularization)

        tracer_normal = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0],
                                                            source_galaxies=[g1],
                                                            image_plane_grid_stack=ld_blur.grid_stack)

        tracer_sensitive = ray_tracing.TracerImageSourcePlanes(
            lens_galaxies=[g0, g0_subhalo], source_galaxies=[g1],
            image_plane_grid_stack=ld_blur.grid_stack)

        fit = sensitivity_fit.SensitivityInversionFit(lens_data=ld_blur,
                                                      tracer_normal=tracer_normal,
                                                      tracer_sensitive=tracer_sensitive)

        assert (fit.fit_normal.image_2d == ld_blur.image_2d).all()
        assert (fit.fit_normal.noise_map_2d == ld_blur.noise_map_2d).all()

        mapper = pixelization.mapper_from_grid_stack_and_border(
            grid_stack=tracer_normal.source_plane.grid_stack,
            border=None)
        inversion = inv.Inversion.from_data_1d_mapper_and_regularization(
            mapper=mapper, regularization=regularization, image_1d=ld_blur.image_1d,
            noise_map_1d=ld_blur.noise_map_1d,
            convolver=ld_blur.convolver_mapping_matrix)

        assert fit.fit_normal.model_image_2d == pytest.approx(
            inversion.reconstructed_data_2d, 1.0e-4)

        residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=ld_blur.image_2d, mask=ld_blur.mask_2d,
            model_data=inversion.reconstructed_data_2d)

        assert fit.fit_normal.residual_map_2d == pytest.approx(residual_map_2d, 1.0e-4)

        chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_2d, mask=ld_blur.mask_2d,
            noise_map=ld_blur.noise_map_2d)

        assert fit.fit_normal.chi_squared_map_2d == pytest.approx(chi_squared_map_2d,
                                                                  1.0e-4)

        assert (fit.fit_sensitive.image_2d == ld_blur.image_2d).all()
        assert (fit.fit_sensitive.noise_map_2d == ld_blur.noise_map_2d).all()

        mapper = pixelization.mapper_from_grid_stack_and_border(
            grid_stack=tracer_sensitive.source_plane.grid_stack,
            border=None)
        inversion = inv.Inversion.from_data_1d_mapper_and_regularization(
            mapper=mapper, regularization=regularization, image_1d=ld_blur.image_1d,
            noise_map_1d=ld_blur.noise_map_1d,
            convolver=ld_blur.convolver_mapping_matrix)

        assert fit.fit_sensitive.model_image_2d == pytest.approx(
            inversion.reconstructed_data_2d, 1.0e-4)

        residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=ld_blur.image_2d, mask=ld_blur.mask_2d,
            model_data=inversion.reconstructed_data_2d)

        assert fit.fit_sensitive.residual_map_2d == pytest.approx(residual_map_2d,
                                                                  1.0e-4)

        chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_2d, mask=ld_blur.mask_2d,
            noise_map=ld_blur.noise_map_2d)

        assert fit.fit_sensitive.chi_squared_map_2d == pytest.approx(chi_squared_map_2d,
                                                                     1.0e-4)

        chi_squared_normal = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=fit.fit_normal.chi_squared_map_2d, mask=ld_blur.mask_2d)
        chi_squared_sensitive = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=fit.fit_sensitive.chi_squared_map_2d, mask=ld_blur.mask_2d)
        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            mask=ld_blur.mask_2d, noise_map=ld_blur.noise_map_2d)

        assert fit.fit_normal.likelihood == -0.5 * (
                chi_squared_normal + noise_normalization)
        assert fit.fit_sensitive.likelihood == -0.5 * (
                chi_squared_sensitive + noise_normalization)

        assert fit.figure_of_merit == fit.fit_sensitive.likelihood - fit.fit_normal.likelihood

        fit_from_factory = sensitivity_fit.fit_lens_data_with_sensitivity_tracers(
            lens_data=ld_blur,
            tracer_normal=tracer_normal,
            tracer_sensitive=tracer_sensitive)

        assert fit.figure_of_merit == fit_from_factory.figure_of_merit
